#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This is the ``esm_archiving`` module."""

# Standard Library:
import logging
import os
import re
import subprocess
import sys
import tarfile

# Third-Party Libraries:
import pandas as pd

# This Library
from esm_archiving.external.pypftp import Pftp


class DatestampLocationError(Exception):
    pass


def query_yes_no(question, default="yes"):  # pragma: no cover
    """Ask a yes/no question via ``input()`` and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.

    It must be "yes" (the default), "no" or None (meaning an answer is
    required of the user).

    The "answer" return value is True for "yes" or False for "no".

    Note: Shamelessly stolen from StackOverflow It's not hard to implement, but
    Paul is lazy...

    Parameters
    ----------
    question : str
        The question you'd like to ask the user
    default : str
        The presumed answer for ``question``. Defaults to "yes".

    Returns
    -------
    bool :
        True if the user said yes, False if the use said no.
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def find_indices_of(char, in_string):
    """
    Finds indicies of a specific character in a string

    Parameters
    ----------
    char : str
        The character to look for
    in_string : str
        The string to look in

    Yields
    ------
    int
        Each round of the generator gives you the next index for the desired
        character.
    """
    index = -1
    while True:
        index = in_string.find(char, index + 1)
        if index == -1:
            break
        yield index


def group_indexes(index_list):
    """
    Splits indexes into tuples of monotonically ascending values.

    Parameters
    ----------
    list :
        The list to split up

    Returns
    -------
    list :
        A list of tuples, so that you can get only one group of ascending
        tuples.

    Example
    -------
    >>> indexes = [0, 1, 2, 3, 12, 13, 15, 16]
    >>> group_indexes(indexes)
    [(0, 1, 2, 3), (12, 13), (15, 16)]
    """
    rlist, group = [], []
    for idx, i in enumerate(index_list):
        group.append(i)
        try:
            if i + 1 != index_list[idx + 1]:
                rlist.append(tuple(group))
                group = []
        except IndexError:
            if idx == len(index_list) - 1:
                if i - 1 == index_list[idx - 1]:
                    rlist.append(tuple(group))
                else:
                    rlist.append((i,))
    return rlist


# PLAN
# Go through the experiment and determine the filetypes
# For each filetype, build file groups
def group_files(top, filetype):
    """
    Generates quasi-regexes for a specific filetype, replacing all numbers with
    #.

    Parameters
    ----------
    top : str
        Where to start looking (this should normally be top of the experiment)
    filetype : str
        Which files to go through (e.g. outdata, restart, etc...)

    Returns
    -------
    dict
        A dictonary containing keys for each folder found in ``filetype``, and
        values as lists of files with strings where numbers are replaced by #.
    """
    model_dirs = _generate_model_dirs(top, filetype)
    model_files = {model: None for model in model_dirs}
    for model_dir in model_dirs:
        cleaned_filepatterns = _walk_through_model_dir_clean_numbers(
            os.path.join(top, filetype, model_dir)
        )
        model_files[model_dir] = cleaned_filepatterns
    return model_files


def _generate_model_dirs(top, filetype):
    model_dirs = [
        model
        for model in os.listdir(os.path.join(top, filetype))
        if os.path.isdir(os.path.join(top, filetype, model))
    ]
    return model_dirs


def _clean_numbers_nonlonely(s):
    characters = list(s)
    number_replacements = []
    for idx, char in enumerate(characters):
        if idx == 0:
            if char.isdigit() and characters[1].isdigit():
                number_replacements.append(idx)
            continue
        while idx < len(characters) - 1:
            prev_char = characters[idx - 1]
            next_char = characters[idx + 1]
            if char.isdigit() and (prev_char.isdigit() or next_char.isdigit()):
                number_replacements.append(idx)
            break
        if idx == len(characters) - 1:
            prev_char = characters[idx - 1]
            if prev_char.isdigit() and char.isdigit():
                number_replacements.append(idx)
    rlist = []
    for idx, char in enumerate(characters):
        if idx in number_replacements:
            rlist.append("#")
        else:
            rlist.append(char)
    return "".join(rlist)


def _walk_through_model_dir_clean_numbers(top, allow_single_numbers=False):
    cleaned_filepatterns = []
    for root, _, files in os.walk(top):
        for file in files:
            if allow_single_numbers:
                filepattern = "".join([r"#" if c.isdigit() else c for c in file])
            else:
                filepattern = _clean_numbers_nonlonely(file)
            filepattern = os.path.join(root, filepattern)
            if filepattern not in cleaned_filepatterns:
                cleaned_filepatterns.append(filepattern)
    return cleaned_filepatterns


def purify_expid_in(model_files, expid, restore=False):
    """
    Puts or restores >>>EXPID<<< marker in filepatterns

    Parameters
    ----------
    model_files : dict
        The model files for archiving
    expid : str
        The experiment ID to purify or restore
    restore : bool
        Set experiment ID back from the temporary marker

    Returns
    -------
    dict :
        Dictionary containing keys for each model, values for file patterns
    """
    for model in model_files:
        for idx, filepattern in enumerate(model_files[model]):
            try:
                if restore:
                    new_filepattern = filepattern.replace(">>>EXPID<<<", expid)
                else:
                    new_filepattern = filepattern.replace(expid, ">>>EXPID<<<")
                model_files[model][idx] = new_filepattern
            except KeyError:
                logging.debug("Can't replace experiment id for: %s", filepattern)
    return model_files


def stamp_files(model_files):
    """
    Given a sttandard file dictioanry (keys: model names, values: filepattern);
    figures out where the date probably is, and replaces the ``#`` sequence
    with a >>>DATE<<< stamp.

    Parameters
    ----------
    model_files : dict
        Dictionary of keys (model names) where values are lists of files for
        each model.

    Returns
    -------
    dict :
        As the input, but replaces the filepatterns with the >>>DATE<<< stamp.
    """
    for model in model_files:
        idx_modifier = 0
        for idx, filepattern in enumerate(model_files[model]):
            idx += idx_modifier
            try:
                stamped_filepattern = stamp_filepattern(filepattern)
                logging.debug(f"Adding [{model}][{idx}] = {stamped_filepattern}")
                model_files[model][idx] = stamped_filepattern
            except AssertionError:  # List was longer than 1
                stamped_filepatterns = stamp_filepattern(filepattern, force_return=True)
                logging.debug(stamped_filepatterns)
                idx_modifier += len(stamped_filepatterns)
                for idx_modifier, stamped_filepattern in enumerate(
                    stamped_filepatterns
                ):
                    idx += idx_modifier
                    logging.debug(f"Adding [{model}][{idx}] = {stamped_filepattern}")
                    model_files[model][idx] = stamped_filepattern
                continue  # needed??
    return model_files


def get_list_from_filepattern(filepattern):
    dirname = os.path.dirname(filepattern)
    regex_files = re.compile(os.path.basename(filepattern).replace("#", "\d"))  # noqa
    matching_files = sorted(
        [
            os.path.join(dirname, file)
            for file in os.listdir(dirname)
            if re.match(regex_files, file)
        ]
    )
    return matching_files


def stamp_filepattern(filepattern, force_return=False):
    """
    Transforms # in filepatterns to >>>DATE<<< and replaces other numbers back
    to original

    Parameters
    ----------
    filepattern : str
        Filepattern to get date stamps for

    force_return : bool
        Returns the list of filepatterns even if it is longer than 1.

    Returns
    -------
    str :
        New filepattern, with >>>DATE<<<
    """
    logging.debug("Incoming filepattern: ", filepattern)
    # If >>>DATE<<< is already in the filepattern, this should be a no-op:
    if ">>>DATE<<<" in filepattern:
        return filepattern
    # No number in this filepattern, nothing to do:
    if "#" not in filepattern:
        return filepattern
    # Get full file list:
    matching_files = get_list_from_filepattern(filepattern)
    try:
        datestamp_location = determine_datestamp_location(matching_files)
    except DatestampLocationError:
        logging.debug("Couldn't determine a unique datestamp for %s" % filepattern)
        return filepattern  # PG: This might be a horrible idea...
    files = [f.replace(f[datestamp_location], ">>>DATE<<<") for f in matching_files]
    files = set(files)
    try:
        assert len(files) == 1
    except AssertionError:
        if force_return:
            return list(files)
        else:
            raise
    filepattern = files.pop()
    return filepattern


# Figure out where the datestamp is in the file pattern
def determine_potential_datestamp_locations(filepattern):
    """
    For a filepattern, gives back index of potential date locations

    Parameters
    ----------
    filepattern : str
        The filepattern to check.

    Returns
    -------
    list :
        A list of slice object which you can use to cut out dates from the
        filepattern
    """
    indexes = list(
        find_indices_of("#", "".join([r"#" if c.isdigit() else c for c in filepattern]))
    )

    aligned_indexes = []
    if indexes[0] + 1 == indexes[1]:
        aligned_indexes.append(0)
    for idx, i in enumerate(indexes[1:-1]):
        idx += 1
        expected_previous = i - 1
        expected_next = i + 1
        if (indexes[idx + 1] == expected_next) or (
            indexes[idx - 1] == expected_previous
        ):
            aligned_indexes.append(i)
    if indexes[-1] - 1 == indexes[-2]:
        aligned_indexes.append(indexes[-1])

    grouped_indexes = group_indexes(aligned_indexes)

    slices = []
    for group in grouped_indexes:
        slices.append(slice(group[0], group[-1] + 1))
    return slices


def determine_datestamp_location(files):
    """
    Given a list of files; figures where the datestamp is by checking if it
    varies.

    Parameters
    ----------
    files : list
        A list (longer than 1!) of files to check

    Returns
    -------
    slice :
        A slice object giving the location of the datestamp

    Raises
    ------
    DatestampLocationError :
        Raised if there is more than one slice found where the numbers vary over
        different files -or- if the length of the file list is not longer than 1.
    """
    try:
        assert len(files) > 1
    except AssertionError:
        raise DatestampLocationError(
            "Unable to determine a datestamp from just 1 file!"
        )
    # Use the first file as a template (Probably a bad idea):
    filepattern = files[0]
    slices = determine_potential_datestamp_locations(filepattern)
    valid_slices = []
    for slice_ in slices:
        if files[0][slice_] == files[1][slice_]:
            continue
        else:
            valid_slices.append(slice_)
    if len(valid_slices) > 1:
        raise DatestampLocationError("Unable to determine a unique datestamp!")
    return valid_slices[0]


# Build a regex to go through a certain number of dates (e.g. 1 decade)
def get_files_for_date_range(
    filepattern, start_date, stop_date, frequency, date_format=r"%Y%m%d"
):
    """
    Creates a list of files for specified start/stop dates

    Parameters
    ----------
    filepattern : str
        A filepattern to replace dates in
    start_date : str
        The starting date, in a pandas-friendly date format
    stop_date : str
        Ending date, pandas friendly. Note that for end dates, you need to
        **add one month** to assure that you get the last step in your list!
    frequency : str
        Frequency of dates, pandas friendly
    date_format : str
        How dates should be formatted, defaults to %Y%m%d

    Returns
    -------
    list
        A list of strings for the filepattern with correct date stamps.

    Example
    -------
    >>> filepattern =  "LGM_24hourly_PMIP4_echam6_BOT_mm_>>>DATE<<<.nc"
    >>> get_files_for_date_range(filepattern, "1890-07", "1891-11", "1M", date_format="%Y%m")
    [
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189007.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189008.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189009.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189010.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189011.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189012.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189101.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189102.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189103.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189104.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189105.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189106.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189107.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189108.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189109.nc",
        "LGM_24hourly_PMIP4_echam6_BOT_mm_189110.nc",
    ]
    """
    # I had initially wanted to do a regex for this; but I get the impression
    # that's a horrible idea. Or at least, a difficult one. If we know the date
    # format, we can just build a list of the desired dates.

    # BUG: determine_datestamp_location takes a list, not a string!!!
    date_stamp = ">>>DATE<<<"
    dates = [
        date.to_pydatetime().strftime(date_format)
        for date in pd.date_range(start=start_date, end=stop_date, freq=frequency)
    ]
    files = [filepattern.replace(date_stamp, str(date)) for date in dates]
    return files


# Sort the files into lists dependening on the dates
def sort_files_to_tarlists(model_files, start_date, end_date, config):
    out_lists = {}
    for model in model_files:
        frequency = config.get(model, {}).get("archive", {}).get("frequency")
        date_format = config.get(model, {}).get("archive", {}).get("date_format")
        out_lists[model] = []
        for filepattern in model_files[model]:
            out_lists[model].append(
                get_files_for_date_range(
                    filepattern,
                    start_date,
                    end_date,
                    frequency,
                    date_format=date_format,
                )
            )
        out_lists[model] = sorted(
            list(set([item for sublist in out_lists[model] for item in sublist]))
        )
    # Remove still empty lists:
    out_lists = {k: v for k, v in out_lists.items() if v}
    return out_lists


# Perform an integrity check
def check_tar_lists(tar_lists):
    existing_tar_lists = {}
    missing_tar_lists = {}
    for model in tar_lists:
        existing_tar_lists[model] = [f for f in tar_lists[model] if os.path.isfile(f)]
        missing_tar_lists[model] = [
            f for f in tar_lists[model] if not os.path.isfile(f)
        ]
    return existing_tar_lists, missing_tar_lists


# Check size of all files in each list
def sum_tar_lists(tar_lists):
    """
    Sums up the amount of space in the tar lists dictionary

    Given ``tar_lists``, which is generally a dicitonary consisting of keys (model
    names) and values (files to be tarred), figures out how much space the
    **raw, uncompressed** files would use. Generally the compressed tarball
    will take up less space.

    Parameters
    ----------
    tar_lists : dict
        Dictionary of file lists to be summed up. Reports every sum as a value
        for the key of that particular list.

    Returns
    -------
    dict :
        Keys are the same as in the input, values are the sums (in bytes) of
        all files present within the list.
    """
    sizes = {}
    for model in tar_lists:
        sizes[model] = sum(os.path.getsize(f) for f in tar_lists[model])
    return sizes


def sum_tar_lists_human_readable(tar_lists):
    """
    As ``sum_tar_lists`` but gives back strings with human-readable sizes.
    """

    def human_readable_size(size, decimal_places=3):
        for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
            if size < 1024.0:
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f}{unit}"

    sizes = {k: human_readable_size(v) for k, v in sum_tar_lists(tar_lists).items()}
    return sizes


# Split into multiple lists based on size:
def split_list_due_to_size_limit(in_list, slimit):
    flist = in_list.copy()
    current_size = 0
    total_list = []
    current_list = []
    while flist:
        current_file = flist.pop(0)
        current_size += os.path.getsize(current_file)
        current_list.append(current_file)
        logging.debug(current_size)
        if current_size >= slimit * 0.8:
            logging.debug("Critical size exceeded! Starting new list!")
            total_list.append(current_list)
            current_size = 0
            current_list = []
    if current_list:
        total_list.append(current_list)
    return total_list


# Utility function to make running commands on the shell easier:
def run_command(command):
    """
    Runs ``command`` and directly prints output to screen.

    Parameters
    ----------
    command : str
        The command to run, with pipes, redirects, whatever

    Returns
    -------
    rc : int
        The return code of the subprocess.
    """
    subprocess.check_call(
        command, stdout=sys.stdout, stderr=subprocess.STDOUT, shell=True
    )


# Pack the files into tarball(s), depending on the size of the list
def pack_tarfile(flist, wdir, outname):
    """
    Creates a compressed tarball (``outname``) with all files found in ``flist``.


    Parameters
    ----------
    flist : list
        A list of files to include in this tarball
    wdir : str
        The directory to "change" to when packing up the tar file. This will
        (essentially) be used in the tar command as the -C option by stripping
        off the beginning of the flist
    outname : str
        The output file name

    Returns
    -------
    str :
        The output file name
    """
    # TODO(pgierz): Would it be more sensible to return the tarball object?
    if not wdir.endswith("/"):
        wdir += "/"
    flist = [item.replace(wdir, "") for item in flist]
    tar_part = (
        f"tar --use-compress-program=pigz -cvf {outname} -C {wdir} {' '.join(flist)}"
    )
    tqdm_part = f"tqdm --total {len(flist)} --unit files"
    output_part = f"{outname}.log"
    run_command(tar_part + "|" + tqdm_part + ">>" + output_part)
    # with tarfile.open(outname, "w:gz") as tar:
    #    for f in tqdm.tqdm(flist):
    #        tar.add(f, arcname=os.path.basename(f))
    return outname


# Write a small log of what is in that tarball
def log_tarfile_contents(tfile):
    """
    Generates a log of the tarball contents

    Parameters
    ----------
    tfile : str
        The path for the tar file to generate a log for

    Returns
    -------
    None

    Warning
    -------
        Note that for this function to work, you need to have write permission in
        the directory where the tarball is located. If not, this will probably
        raise an OSError.  I can imagine giving the location of the log path as
        an argument; but would like to see if that is actually needed before
        implementing it...

    """
    with tarfile.open(tfile) as tar:
        with open(os.path.splitext(tfile)[0] + ".tar_contents", "w") as contents:
            for member in tar.getmembers():
                # INFO: For which attributes you can write, see here:
                # https://docs.python.org/3/library/tarfile.html#tarfile.TarInfo
                for attr in ["name", "size"]:
                    contents.write(str(getattr(member, attr)))
                    contents.write("\t")
                contents.write("\n")


# Upload the tarball to the tape
def archive_mistral(tfile, rtfile=None):
    """
    Puts the ``tfile`` to the tape archive using ``tape_command``

    Parameters
    ----------
    tfile : str
        The full path of the file to put to tape

    rtfile : str
        The filename on the remote tape server. Defaults to None, in which case
        a replacement is performed to keep as much of the filename the same as
        possible. Example: /work/ab0246/a270077/experiment.tgz -->
        /hpss/arch/ab0246/a270077/experiment.tgz


    Returns
    -------
    None
    """
    if not rtfile:
        rtfile = tfile.replace("/work", "/hpss/arch")

    remote_base_dir = os.path.dirname(rtfile)

    tape_server = Pftp()

    if not tape_server.exists(remote_base_dir):
        tape_server.makedirs(remote_base_dir)

    if os.path.isfile(tfile) and os.stat(tfile).st_size > 0:
        # print(archive_name, "will be uploaded to", remote_archive_name)
        tape_server.upload(tfile, rtfile)
    else:
        print(f"{rtfile} doesn't exist or is an empty file, skipping...")


# If requested, delete the original data
def delete_original_data(tfile, force=False):
    """
    Erases data which is found in the tar file.

    Parameters
    ----------
    tfile : str
        Path to the tarfille whose data should be erased.
    force : bool
        If False, asks the user if they really want to delete their files.
        Otherwise just does this silently. Default is ``False``

    Returns
    -------
    None
    """
    with tarfile.open(tfile) as tar:
        if not force:
            print("WARNING! You are about to delete these files:")
            for f in tar.getnames():
                print(f)
            response = query_yes_no("Do you want to continue?", default="no")
            if not response:
                return
        for f in tar.getnames():
            os.remove(f)



from . import database
from datetime import datetime


def database_entry(action, setup_name, base_dir):

    thisrun = (
        database.session.query(database.installation)
        .filter_by(setup_name=setup_name)
        .filter_by(folder=base_dir + "/" + setup_name)
        .all()
    )

    if thisrun == []:
        thisrun = database.installation(
            action=action,
            setup_name=setup_name,
            timestamp=datetime.now(),
            folder=base_dir + "/" + setup_name,
        )
        database.session.add(thisrun)
    else:
        thisrun = thisrun[-1]
        thisrun.action = action
        thisrun.timestamp = datetime.now()
    database.session.commit()
    return database



# -*- coding: utf-8 -*-

"""Top-level package for ESM Archiving."""

__author__ = """Paul Gierz"""
__email__ = "pgierz@awi.de"
__version__ = "6.0.3"

from .esm_archiving import (
    archive_mistral,
    check_tar_lists,
    delete_original_data,
    determine_datestamp_location,
    determine_potential_datestamp_locations,
    find_indices_of,
    get_files_for_date_range,
    get_list_from_filepattern,
    group_files,
    group_indexes,
    log_tarfile_contents,
    pack_tarfile,
    purify_expid_in,
    sort_files_to_tarlists,
    split_list_due_to_size_limit,
    stamp_filepattern,
    stamp_files,
    sum_tar_lists,
    sum_tar_lists_human_readable,
)

__all__ = (
    "archive_mistral",
    "check_tar_lists",
    "delete_original_data",
    "determine_datestamp_location",
    "determine_potential_datestamp_locations",
    "find_indices_of",
    "get_files_for_date_range",
    "get_list_from_filepattern",
    "group_files",
    "group_indexes",
    "log_tarfile_contents",
    "pack_tarfile",
    "purify_expid_in",
    "sort_files_to_tarlists",
    "split_list_due_to_size_limit",
    "stamp_filepattern",
    "stamp_files",
    "sum_tar_lists",
    "sum_tar_lists_human_readable",
)



import sys, os
from . import getch
import datetime


class DisplayDatabase:
    def __init__(self, tablename=None):

        from . import location_database

        query = location_database.session.query(location_database.database_location)
        results = query.all()
        if len(results) == 0:
            print("No tables registered, sorry...")
            sys.exit(-1)

        all_tablenames = [result.table_name for result in results]

        if not tablename:
            pass
        elif tablename not in all_tablenames:
            print("Unknown table name.")
            tablename = None

        if not tablename:
            print("Please choose one of the following tables:")
            for result in results:
                print(result)
            while True:
                choice = input("ID (q to quit): ")
                if choice in ["q", "Q"]:
                    sys.exit(0)
                try:
                    choice = int(choice)
                    tablename = [
                        result.table_name for result in results if result.id == choice
                    ]
                    if tablename == []:
                        print("ID not valid.")
                    else:
                        tablename = str(tablename[0])
                        break
                except:
                    print("ID needs to be of type integer, please try again...")

        print(tablename)

        table = [table for table in results if tablename == table.table_name]
        table = table[0]

        package = __import__(table.class_in)
        database = getattr(package, "database")
        self.entry_type = getattr(database, tablename)

        # sys.exit(0)
        #
        #        if tablename == "experiments":         # that needs to disappear
        ##            from esm_runscripts import database
        #            self.entry_type = database.experiment
        #        else:
        #            print ("Unknown table, quitting...")
        #            sys.exit(-1)
        #
        self.session = database.session
        query = database.session.query(self.entry_type)
        results = query.all()
        if not type(results) == list:
            results = [results]

        self.query = query
        self.verbose = False
        self.all_results = results
        self.results = results
        table = getattr(self.entry_type, "__table__")
        columns = getattr(table, "columns")
        self.columns = [str(column).replace(tablename + ".", "") for column in columns]

        while True:
            self.output_writer()
            self.decision_maker()

    def output_writer(self):
        lines_of_terminal = os.get_terminal_size().lines

        for line in range(0, lines_of_terminal):
            print()
        if not self.verbose:
            getattr(self.entry_type, "topline")()
        for result in self.results:
            if self.verbose:
                getattr(self.entry_type, "nicer_output")(result)
            else:
                print(result)
        print(
            "------------------------------------------------------------------------------------------------------------------------------"
        )
        print(
            "[S]elect Entry        [V]erbose on/off        [R]eset selection        [E]dit Entry        [D]elete Entry               [Q]uit"
        )
        print(
            "------------------------------------------------------------------------------------------------------------------------------"
        )

    def ask_column(self):
        find = input("Which column " + str(self.columns) + "? ").lower()

        if find == "":
            return None

        for column in self.columns:
            if column.startswith(find):
                find_column = column
                break

        if not find_column:
            print("Unknown column.")
            return None

        return find_column

    def ask_dataset(self):
        edit_id = None
        if len(self.results) == 0:
            return None

        elif len(self.results) == 1:
            print("Select entry id=" + str(self.results[0].id) + " [y/n]?")

            yesno = getch.get_one_of(["y", "Y", "n", "N"])
            if yesno.lower() == "y":
                edit_id = self.results[0].id
                edit_dataset = self.results[0]

        if not edit_id:
            edit_id = input("Please enter dataset id: ")
            if edit_id == "":
                return None
            try:
                edit_id = int(edit_id)
            except:
                print("id needs to be of type int.")
                return None
            edit_dataset = [
                result for result in self.all_results if result.id == edit_id
            ]
            if edit_dataset == []:
                print("Unknown dataset id.")
                return None
            else:
                edit_dataset = edit_dataset[0]

        return edit_dataset

    def decision_maker(self):
        char = getch.get_one_of(["q", "Q", "v", "V", "s", "S", "e", "E", "r", "R", "D"])

        if char.lower() == "q":
            sys.exit(0)

        elif char.lower() == "v":
            self.verbose = not self.verbose

        elif char.lower() == "s":
            self.select_stuff()

        elif char.lower() == "e":
            self.edit_dataset()

        elif char.lower() == "r":
            self.results = self.all_results

        elif char == "D":
            self.remove_datasets()

    def remove_datasets(self):
        all_ids = [result.id for result in self.results]
        print("Really remove entry/entries " + str(all_ids) + " [Y/n]? ")
        really = getch.get_one_of(["Y", "n", "N"])
        if really == "Y":
            for this_id in all_ids:
                self.query.filter_by(id=this_id).delete()
            self.session.commit()

            self.all_results = [
                result for result in self.all_results if result.id not in all_ids
            ]
            self.results = self.all_results

    def edit_dataset(self):
        edit_dataset = self.ask_dataset()
        if not edit_dataset:
            return

        edit_column = self.ask_column()
        if not edit_column:
            return
        column_type = type(getattr(self.all_results[0], edit_column))

        edit_entry = input("Change to? ")
        if edit_entry == "":
            return

        # try to convert user input
        if column_type == int:
            try:
                edit_entry = int(edit_entry)
            except:
                print("Wrong data type, should be int!")
                return
        elif column_type == datetime.datetime:
            try:
                edit_entry = datetime.datetime(edit_entry)
            except:
                print("Wrong data type, should be datetime.datetime!")
                return

        this_entry = self.query.filter_by(id=edit_dataset.id).first()
        try:
            setattr(this_entry, edit_column, edit_entry)
            self.session.commit()
        except:
            print(
                "Problem changing column "
                + edit_column
                + " of dataset "
                + str(edit_dataset.id)
                + ". "
            )
            return

        new_results = []
        new_all_results = []
        for result in self.all_results:
            need_this = False
            if result in self.results:
                need_this = True
            if result == edit_dataset:
                setattr(result, edit_column, edit_entry)
            new_all_results += [result]
            if need_this:
                new_results += [result]
        self.results = new_results
        self.all_results = new_all_results

    def select_stuff(self):
        find_column = self.ask_column()
        if not find_column:
            return

        find_value = input("Search for? ")
        if find_value == "":
            return

        if find_column and find_value:
            if find_column == "id":
                self.results = [
                    result
                    for result in self.results
                    if find_value == str(getattr(result, find_column))
                ]
            else:
                self.results = [
                    result
                    for result in self.results
                    if find_value in str(getattr(result, find_column))
                ]



import os
import importlib.util


def subjob_environment(config, subjob):

    task_list = []
    subjob_config = config["general"]["workflow"]["subjobs"][subjob]

    env_preparation = subjob_config.get("env_preparation", False)
    scriptdir = subjob_config.get("script_dir", False)

    if env_preparation:
        env = assemble_filename(env_preparation, scriptdir, config)
        spec = importlib.util.spec_from_file_location(subjob, env)
        envmodule = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(envmodule)

        env_dict = getattr(envmodule, "prepare_environment")(config)
        task_list += export_string(env_dict)

    return task_list


def subjob_tasks(config, subjob, batch_or_shell):

    task_list = []
    subjob_config = config["general"]["workflow"]["subjobs"][subjob]

    old_logfile = config["general"]["logfile_path"]
    logfile_dir = os.path.dirname(old_logfile)
    if config["general"]["setup_name"] in subjob:
        bare_subjob = subjob.replace("_" + config["general"]["setup_name"], "")
    else:
        bare_subjob = subjob
    logfile_name = os.path.basename(old_logfile).replace(
        config["general"]["jobtype"], bare_subjob
    )

    new_logfile = os.path.join(logfile_dir, logfile_name)

    scriptdir = subjob_config.get("script_dir", False)
    script = subjob_config.get("script", False)
    call_function = subjob_config.get("call_function", False)

    if script:
        script = assemble_filename(script, scriptdir, config)
        # task_list += add_scriptcall(script, cluster, config)
        if batch_or_shell == "batch":
            if "calc_launcher_flags" in dir(config['general']["batch"].bs):
                launcher_flags = config['general']["batch"].bs.calc_launcher_flags(
                    {
                        "dataprocess": subjob_config,
                        "computer": config["computer"]
                    },
                    'dataprocess',
                    'pp',
                )
            else:
                launcher_flags = config['computer']['launcher_flags']
            task_list += [
                f"time {config['computer']['launcher']} "
                f"{launcher_flags} {script} 2>&1 &"
            ]
        else:
            if call_function:
                task_list += [". " + script]
                task_list += [call_function + " > " + new_logfile + " 2>&1 &"]
            else:
                task_list += [script + " > " + new_logfile + " 2>&1 &"]

    return task_list


def assemble_filename(filename, dirname, config):
    if filename.startswith("/"):
        return filename
    if filename.startswith(".") or dirname == "." or dirname == "./":
        return os.path.join(["general"]["started_from"], filename)
    if dirname:
        return os.path.join(dirname, filename)
    return os.path.join(["general"]["started_from"], filename)


def export_string(environment_dict):
    export_string = []
    for entry in environment_dict:
        value = environment_dict[entry]
        export_string.append(f"export {entry}={value}")
    return export_string


# def run_job(config):
#    config["general"]["relevant_filetypes"] = ["log", "mon", "outdata", "restart_out","bin", "config", "forcing", "input", "restart_in", "ignore"]
#    helpers.evaluate(config, "dataprocess", "data_recipe")
#    return config
#
#
# def _assemble_dataprocess_tasks(config):
##    """
#    Generates all tasks for data processing which will be written to the run file.
#
#    Parameters
#    ----------
#    data_file
#        File handle to which information should be written.
#
#    Returns
#    -------
#    data_task_list : list
##        The list of post commands which will be executed. These are written
#        to the run file.
#    """
#    datafile = config["general"]["post_file"]
#    data_task_list = []
#
#    # First find the correct subjob_cluster:
#
#    this_cluster = config["general"]["job_type"]
#    config["general"]["this_cluster"] = copy.deepcopy(config["general"]["workflow"]["subjob_clusters"][this_cluster])
##
#    # gather the information on the clusters to be submitted by this job.
#    config["general"]["next_clusters"] = {}
#    for cluster in config["general"]["this_cluster"]["next_submit"]:
#        config["general"]["next_clusters"].update(config["general"]["workflow"]["subjob_clusters"][cluster])
#
##    for subcluster in config["general"]["next_clusters"]:
#        data_task_list.append(tasks_of_one_cluster(config, subcluster))
#
#    return data_task_list
#
#
#
#
# def cluster_scriptcall(scriptcall, cluster, config):
#
#    submit = cluster.get("submit_to_batch_system", False)
#    order = cluster.get("order_in_cluster", False)
#
#    if submit:
#        scriptcall = assemble_srun_command(script)
#
#    if order == "concurrent":
#        scriptcall = scriptcall + ";"
#
#    return scriptcall
#


# def tasks_of_one_subjob(config, cluster, subjob):
#    task_list = []
#    subjob_config = config["general"]["worksflow"]["subjobs"][subjob]
#
#
#    env_preparation = subjob_config.get("env_preparation", False)
#    scriptdir = subjob_config.get("script_dir", False)
#    script = subjob_config.get(script, False)
#
#    if env_preparation:
#        env_preparation = assemble_filename(env_preparation, scriptdir, config)
#        task_list.append(add_environment(env_preparation, config))
#
##    if script:
#        script = assemble_filename(script, scriptdir, config)
#        task_list.append(add_scriptcall(script, cluster, config))
#
#    return task_list
#
#
#
# def tasks_of_one_cluster(config, cluster):
#    task_list = []
#    clusterconfig = config["general"]["next_clusters"][cluster]
#    for subjob in clusterconfig["subjobs"]:
#        task_list.append(tasks_of_one_subjob(config, cluster, subjob))
#
##    return task_list


#    for component in config["general"]["valid_model_names"]:
#        post_file.write(40*"+ "+"\n")
#        post_file.write("Generating post-processing tasks for: %s \n" % component)
#
#        post_task_list.append("\n#Postprocessing %s\n" % component)
#        post_task_list.append("cd "+ config[component]["experiment_outdata_dir"]+"\n")
#
#       pconfig_tasks = config[component].get('postprocess_tasks', {})
##       pconfig_scripts = config[component].get('postprocessing_scripts', {})
#
#        post_file.write("Configuration for post processing: %s \n" % pconfig_tasks)
#
#       for script in pconfig_scripts:
#           postscript_name = pconfig_scripts.get("postprocessing_script_name", None)
##           postscript_dir = pconfig_scripts.get("postprocessing_dir", None)
#
#            envscript_name = pconfig_scripts.get("postprocessing_envscript_name", None)
#
#            postscript_name = assemble_filename(postscript_name, postscript_dir, config)
#           envscript_name = assemble_filename(envscript_name, postscript_dir, config)
##
#            if envscript_name:
#                environment_dict = envscript_name(config)
#                post_task_list += export_string(environment_dict)
#
##            if postscript_name:
#                post_task_list.append(postscript_name)
#
#        post_task_list.append("cd -\n")
#        config["general"]["post_task_list"] = post_task_list
#    return config
#
#
#

# ?????
# def write_simple_postscript(config):
#    batch_system.write_simple_runscript(config)
#    return config



from sqlalchemy import create_engine, Column, Integer, String, Sequence, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

import os

# database_file = os.path.dirname(os.path.abspath(__file__)) + "/../database/esm_runscripts.db"
database_file = os.path.expanduser("~") + "/.esm_tools/esm_runscripts.db"
if not os.path.isdir(os.path.expanduser("~") + "/.esm_tools"):
    os.mkdir(os.path.expanduser("~") + "/.esm_tools")

from esm_database import location_database

engine = create_engine("sqlite:///" + database_file, echo=False)
base = declarative_base()


class experiment(base):
    __tablename__ = "experiment"

    id = Column(Integer, Sequence("user_id_seq"), primary_key=True)
    expid = Column(String)
    timestamp = Column(DateTime)
    runtime = Column(String, default="00:15:00")
    run_timestamp = Column(String, default="00000000-00000000")
    setup_name = Column(String)
    outcome = Column(String, default="crashed")
    cpuh = Column(Integer, default=0)
    gb = Column(Integer, default=0)
    exp_folder = Column(String, default="none yet")
    archive_folder = Column(String, default="none yet")

    location_database.register("experiment", database_file, "esm_runscripts")

    @staticmethod
    def topline():
        print()
        print("EXPERIMENTS:")
        print(
            "{0: >4}".format("ID")
            + "   "
            + "{0: >17}".format("timestamp")
            + "   "
            + "{0: >15}".format("runtime")
            + "   "
            + "{0: >15}".format("expid")
            + "   "
            + "{0: >17}".format("run_time_stamp")
            + "   "
            + "{0: >10}".format("setup")
            + "   "
            + "{0: >5}".format("outcome")
            + "   "
            + "{0: >7}".format("CPUh")
            + "   "
            + "{0: >7}".format("GB")
            + "   "
        )
        print("{0: >126}".format("(this run)   (total)"))

    @staticmethod
    def nicer_output(run):
        print("ID: " + str(run.id) + ", EXPID: " + str(run.expid) + ":")
        print("     Timestamp: " + run.timestamp.strftime("%x %X"))
        print("     Runtime: " + run.runtime)
        print("     Setup: " + run.setup_name)
        print("     Model Run Time: " + str(run.run_timestamp))
        print("     Outcome of run: " + run.outcome)
        print("     Used CPUh of this run: " + str(run.cpuh))
        print("     Used disk space of the whole experiment: " + str(run.gb) + " GB")
        print("     Results in folder: " + run.exp_folder)
        print("     Archived results in folder: " + run.archive_folder)

    def __repr__(self):
        setup = (self.setup_name[:10]) if len(self.setup_name) > 10 else self.setup_name
        return "%4s   %17s   %15s   %15s   %9s   %10s   %7s   %7s   %7s" % (
            str(self.id),
            str(self.timestamp.strftime("%x-%X")),
            str(self.runtime),
            str(self.expid),
            str(self.run_timestamp),
            setup,
            self.outcome,
            str(self.cpuh),
            str(self.gb),
        )


base.metadata.create_all(engine)
connection = engine.connect()
Session = sessionmaker(bind=engine)
Session.configure(bind=engine)
session = Session()



def prepare_environment(config):
    environment_dict = {
            "ICE_TO_FESOM": 0,
            "FESOM_TO_ICE": 1,
            "MESH_DIR_fesom": config["fesom"]["mesh_dir"],
            "xMESH_ROTATED_fesom": config["fesom"]["mesh_rotated"],
            "DATA_DIR_fesom": config["fesom"]["experiment_outdata_dir"],
            "COUPLE_DIR": config["general"]["experiment_couple_dir"],
            "number_of_years_for_forcing": config["model1"]["chunk_size"],
            "CHUNK_SIZE_pism_standalone": config["model2"]["chunk_size"],
            "CHUNK_START_DATE_fesom": config["general"]["chunk_start_date"],
            "CHUNK_END_DATE_fesom": config["general"]["chunk_end_date"],
            "FUNCTION_PATH": "/home/ollie/lackerma/esm_tools/runscripts/awicm-pism/coupling",
            "PYFESOM_PATH": "/home/ollie/lackerma/pyfesom",
            "EXP_ID": config["general"]["command_line_config"]["expid"],
            "iter_coup_regrid_method_ice2oce": "INTERPOLATE",
            "fesom_use_icebergs": config["fesom"]["use_icebergs"],
            "ICEBERG_DIR": config["fesom"]["iceberg_dir"],

            #"FESOM_GRID_input": config["fesom"]["grid_input"],
            #"solidearth_ice_thickness_file":(
            #    config["general"]["experiment_couple_dir"] +
            #    "/ice_thickness.nc"
            #    ),
            #"ADD_UNCHANGED_ICE": config["vilma"].get("add_unchanged_ice", False),
            #"EISLASTFILE_vilma":  (
            #    config["vilma"]["experiment_input_dir"] +
            #    "/" +
            #    config["vilma"]["eislastfile"]
            #    ),
            #"RUN_NUMBER_vilma": config["general"]["run_number"],
            #"RUN_DATE_STAMP": config["general"]["run_datestamp"],
            #"LAST_RUN_DATE_STAMP": config["general"]["last_run_datestamp"],
            #"INITIAL_YEAR_vilma": config["general"]["initial_date"].syear,
            #"NYEAR_vilma_standalone": config["general"]["nyear"],
            #"FINAL_YEAR_vilma": config["general"]["final_date"].syear,
            #"EISLASTCONF_vilma":(
            #    config["vilma"]["experiment_config_dir"] +
            #    "/inp/" +
            #    config["vilma"]["eislastconf"]
            #    )

            }
    
    #if environment_dict["ADD_UNCHANGED_ICE"] == False:
    #    environment_dict["ADD_UNCHANGED_ICE"] = 0
    #elif environment_dict["ADD_UNCHANGED_ICE"] == True:
    #    environment_dict["ADD_UNCHANGED_ICE"] = 1


    print(environment_dict)
    return environment_dict







#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from os import getenv

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "PyGithub",
    "colorama",
    "coloredlogs",
    "emoji",
    "f90nml",
    "gfw-creator",
    "gitpython",
    "loguru",
    "numpy",
    "packaging",
    "pandas>=1.0",
    "psutil",
    "pyyaml",
    "pyyaml>=5.1",
    "questionary",
    "semver",
    "six",
    "sqlalchemy",
    "tabulate",
    "tqdm",
    "typing_extensions>=3.10.0.0",
    "xdgenvpy",
]

setup_requirements = []

test_requirements = ["pyfakefs"]

setup(
    author="Dirk Barbi",
    author_email="dirk.barbi@awi.de",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="ESM Tools external infrastructure for Earth System Modelling",
    entry_points={
        "console_scripts": [
            "esm_archive=esm_archiving.cli:main",
            "esm_cleanup=esm_cleanup.cli:main",
            "esm_database=esm_database.cli:main",
            "esm_master=esm_master.cli:main",
            "esm_plugins=esm_plugin_manager.cli:main",
            "esm_runscripts=esm_runscripts.cli:main",
            "esm_tests=esm_tests.cli:main",
            "esm_tools=esm_tools.cli:main",
            "esm_utilities=esm_utilities.cli:main",
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v2",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="esm_tools",
    name="esm-tools",
    packages=find_packages("src")
    + ["esm_tools", "esm_tools.configs", "esm_tools.namelists", "esm_tools.runscripts"],
    package_dir={
        "": "src",
        "esm_tools.configs": "configs",
        "esm_tools.namelists": "namelists",
        "esm_tools.runscripts": "runscripts",
    },
    package_data={
        "esm_tools.configs": ["../configs/*"],
        "esm_tools.namelists": ["../namelists/*"],
        "esm_tools.runscripts": ["../runscripts/*"],
    },
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/esm-tools/esm_tools",
    version="6.0.3",
    zip_safe=False,
)



#!/usr/bin/env python
# import fileinput, os, sys, getopt

import subprocess
import sys
import os
import yaml

from . import database_actions

from .general_stuff import (
    GeneralInfos,
    version_control_infos,
    tab_completion,
    write_minimal_user_config,
    ESM_MASTER_DIR,
)

from .compile_info import setup_and_model_infos

from .task import Task


def main_flow(parsed_args, target):

    # Creates an object of the class GeneralInfos
    main_infos = GeneralInfos(parsed_args)
    vcs = version_control_infos(parsed_args)

    setups2models = setup_and_model_infos(vcs, main_infos, parsed_args)
    tab_completion(parsed_args, setups2models)
    setups2models.config = setups2models.reduce(target)

    user_config = write_minimal_user_config(setups2models.config)
    # Miguel: Move this somewhere else after talking to Paul and Dirk
    user_config["general"]["run_or_compile"] = "compiletime"

    # deniz: verbose is supposed to be a boolean right? It is initialized as
    # 0 in cli.py. Is it then a debug_level?
    if parsed_args.get("verbose", False):
        user_config["general"]["verbose"] = True

    # kh 27.11.20
    if "modify" in parsed_args:
        if "general" in user_config:
            user_config["general"]["modify_config_file"] = parsed_args["modify"]

    if "ignore" in parsed_args:
        ignore_errors = parsed_args["ignore"]
    else:
        ignore_errors = False

    from esm_runscripts.sim_objects import SimulationSetup

    complete_setup = SimulationSetup(user_config=user_config)
    complete_config = complete_setup.config

    setups2models.update_relevant_entries_with_config(complete_config)

    # This will be a problem later with GEOMAR
    # setups2models.replace_last_vars(env)

    # PG: multi-cluster
    # This is probably not the best name for this...
    #
    # Also note, stuff like recomp probably won't work correctly:
    # $ esm_master recomp-awiesm-2.2/pism
    multi_cluster_job = complete_config.get("general", {}).get("multi_cluster_job")
    if multi_cluster_job:
        original_target = target
        original_task = original_target.split("-")[0]
        original_setup = "-".join(original_target.split("-")[1:])
        os.makedirs(original_setup, exist_ok=True)
        os.chdir(original_setup)
        for realm in multi_cluster_job:
            os.makedirs(realm, exist_ok=True)
            os.chdir(realm)
            subprocess.check_call(
                f"esm_master {original_task}-{multi_cluster_job[realm]}", shell=True
            )
            os.chdir("..")
        return 0

    user_task = Task(
        target, setups2models, vcs, main_infos, complete_config, parsed_args
    )
    if parsed_args.get("verbose", False):
        user_task.output()

    user_task.output_steps()

    if parsed_args.get("check", False):
        # deniz: if the environment variable ESM_MASTER_DEBUG is also set dump
        # the contents of the current config to stdout for more investigation
        if os.environ.get("ESM_MASTER_DEBUG", None):
            print()
            print("Contents of the complete_config:")
            print("--------------------------------")
            print(yaml.dump(complete_config, default_flow_style=False, indent=4))

        print("esm_master: check mode is activated. Not executing the actions above")
        return 0

    user_task.validate()

    user_task.execute(ignore_errors)  # env)

    database = database_actions.database_entry(
        user_task.todo, user_task.package.raw_name, ESM_MASTER_DIR
    )
    database.connection.close()

    if not parsed_args["keep"]:
        user_task.cleanup_script()

    return 0



"""
The database module for archiving.

The database extension allows you keep track of which experiments are on the
tape, which files are in which tarball, along with some experiment meta-data.
"""

from datetime import datetime

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean
from sqlalchemy.orm import relationship

Base = declarative_base()


class Experiments(Base):
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True)
    expid = Column(String, unique=True)
    # One to One relationship with archive:
    archive = relationship("Archive", uselist=False, back_populates="exp_ref")
    created_at = Column(DateTime, default=datetime.utcnow)

    def __repr__(self):
        str_created_at = self.created_at.strftime("%Y-%m-%d %H:%M:%S")
        return "<Image (id='%s', expid='%s', archive='%s', created_at=%s)>" % (
            self.id,
            self.expid,
            self.archive,
            str_created_at,
        )


class Archive(Base):
    __tablename__ = "archives"
    id = Column(Integer, primary_key=True)
    # One to One relationship with expid:
    expid_id = Column(Integer, ForeignKey("experiments.id"))
    exp_ref = relationship("Experiments", back_populates="archive")

    # One to Many relationship with Tarball:
    tarballs = relationship("Tarball", back_populates="archive")


class Tarball(Base):
    __tablename__ = "tarballs"
    id = Column(Integer, primary_key=True)
    fname = Column(String, unique=True)

    # Many to One relationship with Archive
    archive_id = Column(Integer, ForeignKey("archives.id"))
    archive = relationship("Archive", back_populates="tarballs")

    # One to Many relationship with Files
    files = relationship("ArchivedFile", back_populates="tarball")


class ArchivedFile(Base):
    __tablename__ = "files"
    id = Column(Integer, primary_key=True)
    tarball_id = Column(Integer, ForeignKey("tarballs.id"))
    tarball = relationship("Tarball", back_populates="files")
    fname = Column(String, unique=True)
    on_tape = Column(Boolean, default=False)
    on_disk = Column(Boolean, default=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

# -*- coding: utf-8 -*-

# ppftp.py

"""
python interface to HPSS's pftp

"""

import os
import re
import sys
import time
import netrc
import ftplib
import getpass
import fnmatch
import posixpath
import subprocess as sp

from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor


HOST = "tape.dkrz.de"
PORT = 4021


__all__ = ["upload", "download", "Pftp"]


def connect(username=None, password=None):

    ftp_obj = ftplib.FTP()
    ftp_obj.connect(HOST, PORT)

    username = username or getpass.getuser()
    if password is None:
        try:
            n = netrc.netrc()
            username, account, password = n.authenticators(HOST)
        except netrc.NetrcParseError as e:
            print("Could not parse .netrc file: %s", e)
            exit(1)
        except IOError as e:
            print("I/O error: {0}".format(e))
            exit(1)

    ftp_obj.login(username, password)
    return ftp_obj


_msg_re = re.compile("^[0-9]+-?\s?[A-Za-z].*").match


def filter_msg(lines):
    try:
        lines = lines.splitlines()
    except AttributeError:
        pass
    for line in lines:
        if line and not _msg_re(line):
            yield line


def msgs(lines):
    try:
        lines = lines.splitlines()
    except AttributeError:
        pass
    for line in lines:
        if _msg_re(line):
            yield line


class Pftp(object):
    HOST = HOST
    PORT = PORT

    def __init__(self, username=None, password=None):
        self.username = username = username or getpass.getuser()
        self.ftp_obj = connect(username, password)
        self._auth = {"username": username, "password": password}

    def __repr__(self):
        is_active = self.is_connected()
        status = "alive" if is_active else "stale"
        s = (
            "<{0.__class__.__name__} connected to "
            "{0.HOST}:{0.PORT} "
            "as {0.username}>"
        )
        s = s.format(self)
        s += "; connection={}".format(status)
        return repr(s)

    def is_connected(self):
        "check if the connection is still active"
        try:
            self.ftp_obj.voidcmd("NOOP")
        except:
            return False
        return True

    def reconnect(self):
        "reconnects to the ftp server"
        try:
            self.ftp_obj.close()
        except:
            pass
        self.ftp_obj = connect(self._auth["username"], self._auth["password"])

    def quit(self):
        self.ftp_obj.quit()

    def close(self):
        self.quit()
        self.ftp_obj.close()

    def pwd(self):
        "present working directory"
        return self.ftp_obj.pwd()

    def cwd(self, path):
        "change working directory"
        try:
            if path.startswith(self.pwd()):
                path = path.replace(self.pwd(), "").lstrip("/")
            self.ftp_obj.cwd(path)
        except:
            raise ValueError("Path does not exists: " + path)

    def exists(self, path):
        "check if a path exists"
        try:
            result = self.stat(path)
        except:
            result = None
        return True if result else False

    def stat(self, pathname):
        "Returns stat of the path"
        try:
            _stat = self.ftp_obj.sendcmd("STAT {}".format(pathname))
        except ftplib.error_perm:
            return None
        return list(filter_msg(_stat))

    def size(self, pathname):
        "Returns size of path in bytes"
        return int(self.stat(pathname)[0].split()[6])

    def isfile(self, pathname):
        "Returns true if pathname refers to an existing file"
        _stat = self.stat(pathname)
        if _stat is None:
            raise ValueError("path does not exists %s" % pathname)
        line = _stat.pop(0)
        if line[0] == "-":
            return True
        if line[0] == "l":
            pwd = self.pwd()
            name = line.split()[-1]
            _flag = False
            try:
                self.cwd(name)
            except:
                _flag = True
            finally:
                self.cwd(pwd)
            return _flag
        return False

    def isdir(self, pathname):
        "Returns true if pathname refers to an existing directory"
        _stat = self.stat(pathname)
        if _stat is None:
            raise ValueError("path does not exists %s" % pathname)
        line = _stat.pop(0)
        if line[0] == "d":
            return True
        if line[0] == "l":
            pwd = self.pwd()
            name = line.split()[-1]
            try:
                self.cwd(name)
                _flag = True
            except ftplib.error_perm:
                _flag = False
            finally:
                self.cwd(pwd)
            return _flag
        return False

    def islink(self, pathname):
        _stat = self.stat(pathname)
        if _stat is None:
            raise ValueError("path does not exists %s" % pathname)
        line = _stat.pop(0)
        if line[0] == "l":
            return True
        return False

    def listing(self, path=None):
        "list directory contents"
        path = path or self.pwd()
        return self.ftp_obj.nlst(path)

    listdir = listing

    def listing2(self, path=None):
        'directory listing in long form. similar to "ls -l"'
        path = path or self.pwd()
        tmp = []
        self.ftp_obj.dir(path, tmp.append)
        return tmp

    def mlsd(self, path):
        cmd = "MLSD %s" % path
        lines = []
        self.ftp_obj.retrlines(cmd, lines.append)
        CLRF = self.ftp_obj
        for line in lines:
            facts, _, name = line.rstrip().partition(" ")
            entry = {}
            for fact in facts[:-1].split(";"):
                key, _, value = fact.partition("=")
                entry[key.lower()] = value
            yield (name, entry)

    def files(self, path=None):
        "gather files at the given path"
        path = path or self.pwd()
        result = []
        for entry in self.listing2(path):
            name = entry.split()[-1]
            if entry.startswith("-"):
                result.append(name)
            if entry.startswith("l"):
                if self.isfile(name):
                    result.append(name)
        return result

    def directories(self, path=None):
        "gather directories at the given path"
        path = path or self.pwd()
        result = []
        for entry in self.listing2(path):
            name = entry.split()[-1]
            if entry.startswith("d"):
                result.append(name)
            if entry.startswith("l"):
                if self.isdir(name):
                    result.append(name)
        return result

    def walk(self, path=None):
        "recursively walk the directory tree from the given path. Similar to os.walk"
        path = path or self.pwd()
        _files = self.files(path)
        _subdirs = self.directories(path)
        yield (path, _subdirs, _files)
        for subdir in _subdirs:
            for x in self.walk(subdir):
                yield x

    def walk_for_files(self, path=None):
        "recursively gather files"
        for root, _dirs, _files in self.walk(path):
            for f in _files:
                yield f

    def walk_for_directories(self, path=None):
        "recursively gather directories"
        for root, _dirs, _files in self.walk(path):
            for d in _dirs:
                yield d

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    # ========== MANIPULATIONS ==========

    def mkdir(self, path):
        if not self.exists(path):
            print("Created directory: " + path)
            self.ftp_obj.mkd(path)

    def makedirs(self, path):
        """Recursively create dirs as required walking up to an existing
        parent dir"""
        try:
            _isdir = self.isdir(path)
        except ValueError:
            pass
        else:
            if _isdir:
                return
        dirname = posixpath.dirname(path)
        if dirname:
            self.makedirs(dirname)
        self.mkdir(path)
        return

    def rmdir(self, path):
        "remove directory"
        if self.exists(path):
            self.ftp_obj.rmd(path)
            print("Removed directory: " + path)
        return

    def remove(self, filename):
        if self.exists(filename):
            self.ftp_obj.delete(filename)
            print("Deleted file: " + filename)
        return

    def removedirs(self, path):
        remove = self.remove
        rmdir = self.rmdir
        path = posixpath.normpath(path)
        dirs = set()
        for topdir, subdirs, files in self.walk(path):
            for _file in files:
                remove(_file)
            if subdirs:
                dirs.update(subdirs)
            dirs.add(topdir)
        dirs = sorted(dirs, key=lambda x: len(x.split(posixpath.sep)), reverse=True)
        for d in dirs:
            rmdir(d)
        return None

    def rename(self, from_name, to_name):
        if os.path.dirname(to_name) and (
            os.path.dirname(from_name) != os.path.dirname(to_name)
        ):
            raise ValueError("Must rename a file within the same directory")
        self.ftp_obj.rename(from_name, to_name)
        return to_name

    # ========== FILE-TRANSFERS ==========

    @staticmethod
    def upload(source, destination):
        "uses pftp binary for transfering the file"
        return upload(source, destination)

    @staticmethod
    def download(source, destination):
        "uses pftp binary for transfering the file"
        return download(source, destination)


def printProgressBar(
    iteration, total, prefix="", suffix="", decimals=1, length=80, fill=u"+"
):
    "https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console"
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print("\r%s |%s| %s%% %s" % (prefix, bar, percent, suffix), end="\r")
    # Print New Line on Complete
    # if iteration == total:
    #    print()
    return


def _transfer(cmdlines):
    client = sp.Popen("pftp", stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
    o, e = client.communicate(cmdlines.encode("utf-8"))
    # for line in o:
    #    print(line)
    return


def _file_stripe_count(filename):
    cmd = ["lfs", "getstripe", "-c", filename]
    return int(sp.check_output(cmd))


def _pftp_upload_commands(source, destination):
    source_filename = os.path.basename(source)
    source_dirname = os.path.dirname(source) or os.curdir
    cmd = ["prompt"]
    stripe_count = _file_stripe_count(source)
    if stripe_count > 1:
        if stripe_count > 8:
            stripe_count = 8  # limit to 8 parallel tasks
        cmd.append("setpwidth {}".format(stripe_count))
        cmd.append("setpblocksize {}".format(1024 * 1024))  # 1MB
    if source_dirname != os.curdir:
        cmd.append("lcd {}".format(source_dirname))
    cmd.append("cd {}".format(destination))
    cmd.append("put {}".format(source_filename))
    cmd.append("bye")
    cmd = "\n".join(cmd)
    return cmd


def upload(source, destination):
    source = source.rstrip("/")
    destination = destination.rstrip("/")
    source_filename = os.path.basename(source)
    destination_maybe_filename = os.path.basename(destination)
    destination_dirname = os.path.dirname(destination)

    p = Pftp()

    requires_renaming = False
    if not p.exists(destination):
        if source_filename == destination_maybe_filename:
            destination = destination_dirname
        elif os.path.splitext(destination_maybe_filename)[-1]:
            destination = destination_dirname
            requires_renaming = True

    destination_filepath = os.path.join(destination, source_filename)

    if not p.exists(destination):
        p.makedirs(destination)

    source_filesize = os.stat(source).st_size

    def get_filesize(filepath):
        if not p.is_connected():
            p.reconnect()
        if not p.exists(filepath):
            return 0
        return p.size(filepath)

    cmdlines = _pftp_upload_commands(source, destination)
    pool = ProcessPoolExecutor(1)
    fut = pool.submit(_transfer, cmdlines)

    # print('Uploading file: {}'.format(source_filename))
    while fut.running():
        time.sleep(1)
        current_size = get_filesize(destination_filepath)
        # if not current_size:
        #    print('\rWaiting for tape to respond...', end='\r')
        # else:
        #    printProgressBar(current_size, source_filesize, source_filename)
        printProgressBar(current_size, source_filesize, source_filename)

    fut.result()
    if requires_renaming:
        p.cwd(destination_dirname)
        p.rename(source_filename, destination_maybe_filename)

    try:
        p.close()
    except:
        pass

    if requires_renaming:
        return os.path.join(destination, destination_maybe_filename)
    return os.path.join(destination, source_filename)


def _pftp_download_commands(source, destination):
    source_filename = os.path.basename(source)
    source_dirname = os.path.dirname(source) or os.curdir
    cmd = ["prompt"]
    cmd.append("cd {}".format(source_dirname))
    if source_dirname != os.curdir:
        cmd.append("lcd {}".format(destination))
    cmd.append("get {}".format(source_filename))
    cmd.append("bye")
    cmd = "\n".join(cmd)
    return cmd


def download(source, destination):
    source = source.rstrip("/")
    destination = destination.rstrip("/")
    source_filename = os.path.basename(source)
    destination_maybe_filename = os.path.basename(destination)
    destination_dirname = os.path.dirname(destination)

    requires_renaming = False
    if not os.path.exists(destination):
        if source_filename == destination_maybe_filename:
            destination = destination_dirname
        elif os.path.splitext(destination_maybe_filename)[-1]:
            destination = destination_dirname
            requires_renaming = True

    destination_filepath = os.path.join(destination, source_filename)

    if not os.path.exists(destination):
        try:
            os.makedirs(destination)
        except:
            pass

    with Pftp() as p:
        source_filesize = p.size(source)

    def get_filesize(filepath):
        if not os.path.exists(filepath):
            return 0
        return os.path.getsize(filepath)

    cmdlines = _pftp_download_commands(source, destination)
    pool = ProcessPoolExecutor(1)
    fut = pool.submit(_transfer, cmdlines)

    # print('Downloading file: {}'.format(source_filename))
    while fut.running():
        time.sleep(1)
        current_size = get_filesize(destination_filepath)
        printProgressBar(
            current_size, source_filesize, os.path.basename(destination_filepath)
        )

    fut.result()
    if requires_renaming:
        curdir = os.path.abspath(os.curdir)
        os.chdir(destination_dirname)
        os.rename(source_filename, destination_maybe_filename)
        os.chdir(curdir)

    if requires_renaming:
        return os.path.join(destination, destination_maybe_filename)
    return os.path.join(destination, source_filename)


def command_line_interface():
    import click

    wildcard_check = re.compile("[*?[]")

    def has_wildcard(path):
        return wildcard_check.search(path) is not None

    class WildcardInDirectoryPath(click.ParamType):
        name = "PathWildcard"

        def convert(self, value, param, ctx):
            try:
                dirname = os.path.dirname(value)
                if has_wildcard(dirname):
                    raise ValueError
                return value
            except ValueError:
                msg = "Wildcard only in file names are allowed {}".format(value)
                self.fail(msg, param, ctx)

    WILDCARD_PATH = WildcardInDirectoryPath()

    def sizeof_fmt(num, suffix="B"):
        for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
            if abs(num) < 1024.0:
                return "%3.1f%s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f%s%s" % (num, "Yi", suffix)

    @click.group()
    def main():
        """A command line interface for HPSS (tape archive)."""

    @main.command("download", short_help="Download files from tape")
    @click.argument("src", nargs=-1, type=WILDCARD_PATH)
    @click.argument("dst", nargs=1, type=click.Path())
    def _download(src, dst):
        """Download files from tape.

        \b
        SRC: file on tape
        DST: path on mistral
        """
        p = Pftp()
        files = []
        for entry in src:
            basename = os.path.basename(entry)
            if has_wildcard(basename):
                dirname = os.path.dirname(entry)
                contents = p.listdir(dirname)
                for fullname in contents:
                    if fnmatch.fnmatch(os.path.basename(fullname), basename):
                        files.append(fullname)
            else:
                if p.exists(entry):
                    files.append(entry)
        total_size = sum(map(p.size, files))
        nfiles = len(files)
        p.close()

        click.echo(
            "{} files to download. Total size: {}".format(
                nfiles, sizeof_fmt(total_size)
            )
        )
        width = str(len(str(nfiles)))
        msg = "[{:>" + width + "}/{:>" + width + "}]"
        for index, fname in enumerate(files, 1):
            downloaded_file = download(fname, dst)
            count = msg.format(index, nfiles)
            click.echo(count + " Downloaded: " + downloaded_file + "  ")

    @main.command("upload", short_help="Upload files to tape")
    @click.argument("src", nargs=-1, type=click.Path())
    @click.argument("dst", nargs=1, type=click.Path())
    def _upload(src, dst):
        """Upload files to tape.

        \b
        SRC: file on mistral
        DST: path on tape
        """
        nfiles = len(src)
        total_size = sum(map(os.path.getsize, src))
        width = str(len(str(nfiles)))
        msg = "[{:>" + width + "}/{:>" + width + "}]"
        click.echo(
            "{} files to upload. Total size: {}".format(nfiles, sizeof_fmt(total_size))
        )
        for index, fname in enumerate(src, 1):
            uploaded_file = upload(fname, dst)
            count = msg.format(index, nfiles)
            click.echo(count + " Uploaded: " + uploaded_file + "  ")

    '''
    @main.command()
    @click.option('-l', default=False, is_flag=True, help="long listing")
    @click.argument('path', type=WILDCARD_PATH)
    def ls(path, l):
        """List directory contents on tape.

        \b
        PATH: path on tape
        """
        p = Pftp()
        if has_wildcard(path):
            listings = p.listdir(os.path.dirname(path))
            files = fnmatch.filter(listings, path)
            if l:
                listings2 = p.listing2(os.path.dirname(path))
                files = [long_path for short_path, long_path in zip(listings, listings2)
                         if short_path in files]
            for fname in files:
                click.echo(fname)
        else:
            if not p.exists(path):
                raise ValueError('Path does not exists')
                # click.echo('Path not found...')
            if l:
                contents = p.listing2(path)
            else:
                contents = p.listdir(path)
            for line in contents:
                click.echo(line)
        p.close()
        return
    '''

    @main.command()
    @click.option("-l", default=False, is_flag=True, help="long listing")
    @click.option("-r", default=False, is_flag=True, help="reverse listing")
    @click.option("-t", default=False, is_flag=True, help="sort by mtime")
    @click.option("-h", default=False, is_flag=True, help="human readable size")
    @click.argument("path", type=click.Path())
    def ls(path, l, r, t, h):
        """List directory contents on tape.

        \b
        PATH: path on tape
        """

        def sanitize_path(path):
            return path.split(None, 8)[-1].split(None)[0]

        p = Pftp()
        result = p.listing2(path)
        aux = {}
        names = dict(zip(list(map(sanitize_path, result)), result))
        paths = defaultdict(list)
        for name in names:
            paths[os.path.dirname(name)].append(name)
        for dname, _files in paths.items():
            mlist = dict(list(p.mlsd(dname)))
            for ff in _files:
                entry = mlist[ff]
                aux[names[ff]] = int(entry.get("size", 0)), entry.get("modify", "")
        if t:
            result = sorted(result, key=lambda row: aux[row][-1])
        if not r:
            result = list(reversed(result))
        if h:
            tmp = []
            for line in result:
                size = aux[line][0]
                parts = line.split(None, 8)
                parts[4] = sizeof_fmt(size)
                tmp.append("  ".join(parts))
            result = tmp[:]
        if not l:
            result = [row.split(None, 8)[-1] for row in result]
        for row in result:
            click.echo(row)
        p.close()
        return

    @main.command()
    @click.argument("path")
    def exists(path):
        """check if PATH exists on tape.

        \b
        PATH: path on tape
        """
        with Pftp() as p:
            if not p.exists(path):
                click.echo("False")
            else:
                click.echo("True")

    @main.command()
    @click.option("-h", is_flag=True, help="human readable format")
    @click.argument("path")
    def size(path, h):
        """Prints file size in bytes. Use -h option for human readable format.

        \b
        PATH: path on tape
        """

        with Pftp() as p:
            if not p.exists(path):
                click.echo("File not found")
            else:
                fsize = p.size(path)
                if h:
                    fsize = sizeof_fmt(fsize)
                click.echo("{}".format(fsize))

    @main.command()
    @click.argument("path")
    def mkdir(path):
        """creates PATH on tape.

        \b
        PATH: path on tape
        """
        with Pftp() as p:
            if p.exists(path):
                click.echo("Already exists")
            else:
                p.makedirs(path)

    @main.command()
    @click.argument("path")
    def isdir(path):
        """checks if PATH is a directory.

        \b
        PATH: path on tape
        """
        with Pftp() as p:
            if not p.exists(path):
                click.echo("path not found")
            elif p.isdir(path):
                click.echo("True")
            else:
                click.echo("False")

    @main.command()
    @click.argument("path")
    def isfile(path):
        """checks if PATH is a file.

        \b
        PATH: path on tape
        """
        res = 1
        with Pftp() as p:
            if not p.exists(path):
                click.echo("path not found")
            elif p.isfile(path):
                click.echo("True")
                res = 0
            else:
                click.echo("False")
        sys.exit(res)

    @main.command()
    @click.argument("path")
    def islink(path):
        """checks if PATH is a link.

        \b
        PATH: path on tape
        """
        with Pftp() as p:
            if not p.exists(path):
                click.echo("path not found")
            elif p.islink(path):
                click.echo("True")
            else:
                click.echo("False")

    def pyversion_check():
        import os

        supported_versions = ("python/3.5.2", "python/2.7.12")
        modules = os.environ.get("LOADEDMODULES", "")
        for version in supported_versions:
            if version in modules:
                break
        else:
            pyversion = "".join([v for v in modules.split(":") if "python" in v])
            msg = """\nLOADED VERSION: {}
SUPPORTED VERSIONS: {}
USE "module load" COMMAND TO LOAD A SUPPORTED PYTHON VERSION.
""".format(
                pyversion, " or ".join(supported_versions)
            )
            import warnings

            warnings.warn(msg, RuntimeWarning, stacklevel=2)

    # pyversion_check()
    main()


if __name__ == "__main__":
    command_line_interface()
